__title__ = "pbipy"
__description__ = "A Python Library for interacting with the Power BI Rest API."
__url__ = "https://github.com/andrewvillazon/pbipy"
__version__ = "2.10.0"
__author__ = "Andrew Villazon"
__author_email__ = "andrew.villazon@gmail.com"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2024 Andrew Villazon"